///////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2020 Martin Wölzer
//
// This code is licensed under the MIT License (MIT).
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
///////////////////////////////////////////////////////////////////////////////

#ifndef CLAPP_VALUE_HPP
#define CLAPP_VALUE_HPP

#include <clapp/contract.h>
#include <clapp/exception.h>
#include <clapp/type_traits.h>
#include <clapp/value.h>

#include <sstream>

template <typename T>
clapp::value::default_value_t<T>::default_value_t(T _value)
    : value{std::move(_value)} {}

template <typename T>
constexpr const char* clapp::value::get_chrono_postfix() noexcept {
    const char* postfix{""};
    if constexpr (std::is_same_v<T, std::chrono::nanoseconds>) {
        postfix = "ns";
    } else if constexpr (std::is_same_v<T, std::chrono::microseconds>) {
        postfix = "µs";
    } else if constexpr (std::is_same_v<T, std::chrono::milliseconds>) {
        postfix = "ms";
    } else if constexpr (std::is_same_v<T, std::chrono::seconds>) {
        postfix = "s";
    } else if constexpr (std::is_same_v<T, std::chrono::minutes>) {
        postfix = "m";
    } else if constexpr (std::is_same_v<T, std::chrono::hours>) {
        postfix = "h";
    }
    return postfix;
}

template <typename T>
std::string clapp::value::to_string(const T& value) {
    std::stringstream string_stream;
    if constexpr (std::is_integral_v<T>) {
        string_stream << std::to_string(value);
    } else if constexpr (type_traits::is_chrono_duration<T>::value) {
        string_stream << value.count() << get_chrono_postfix<T>();
    } else {
        string_stream << value;
    }
    return string_stream.str();
}

template <typename T>
std::string clapp::value::default_value_t<T>::append_restriction_text() const {
    const std::string val_str{to_string(value)};
    return "default-value: " + val_str;
}

template <typename T>
T clapp::value::default_value_t<T>::default_value() const {
    return value;
}

template <typename T>
clapp::value::min_max_value_t<T>::min_max_value_t(T _min, T _max)
    : min{_min}, max{_max} {
    CLAPP_CONTRACT_EXPECTS(min < max);
}

template <typename T>
std::string clapp::value::min_max_value_t<T>::append_restriction_text() const {
    const std::string min_str{to_string(min)};
    const std::string max_str{to_string(max)};
    return "constraint: [" + min_str + "," + max_str + "]";
}

template <typename T>
void clapp::value::min_max_value_t<T>::validate(
    const T& value, const std::string& param_name) const {
    if (value < min) {
        throw clapp::out_of_range_t{"CLI value for '" + param_name +
                                    "' is too low (" + to_string(value) +
                                    " < " + to_string(min) + ")"};
    }
    if (value > max) {
        throw clapp::out_of_range_t{"CLI value for '" + param_name +
                                    "' is too high (" + to_string(value) +
                                    " > " + to_string(max) + ")"};
    }
}

template <typename T>
clapp::value::not_null_value_t<T>::not_null_value_t() = default;

template <typename T>
constexpr std::string_view
clapp::value::not_null_value_t<T>::append_restriction_text() noexcept {
    return "constraint: not null";
}

template <typename T>
void clapp::value::not_null_value_t<T>::validate(
    const T& value, const std::string& param_name) const {
    if (value == T{0}) {
        throw clapp::out_of_range_t{"CLI value for '" + param_name +
                                    "' must not be null (" + to_string(value) +
                                    ")"};
    }
}

constexpr clapp::value::exit_t::exit_t(const int exit_code_param) noexcept
    : exit_code{exit_code_param} {}

constexpr clapp::value::exit_t clapp::value::exit_t::exit(
    const int exit_code_param) noexcept {
    return exit_t{exit_code_param};
}

constexpr int clapp::value::exit_t::get_exit_code() const noexcept {
    return exit_code;
}

constexpr std::string_view
clapp::value::path_exists_t::append_restriction_text() noexcept {
    return "existing path";
}

template <>
std::string clapp::value::convert_value<std::string>(std::string_view param);

template <>
bool clapp::value::convert_value<bool>(std::string_view param);

template <>
std::int8_t clapp::value::convert_value<std::int8_t>(std::string_view param);

template <>
std::int16_t clapp::value::convert_value<std::int16_t>(std::string_view param);

template <>
std::int32_t clapp::value::convert_value<std::int32_t>(std::string_view param);

template <>
std::int64_t clapp::value::convert_value<std::int64_t>(std::string_view param);

template <>
std::uint8_t clapp::value::convert_value<std::uint8_t>(std::string_view param);

template <>
std::uint16_t clapp::value::convert_value<std::uint16_t>(
    std::string_view param);

template <>
std::uint32_t clapp::value::convert_value<std::uint32_t>(
    std::string_view param);

template <>
std::uint64_t clapp::value::convert_value<std::uint64_t>(
    std::string_view param);

template <>
std::chrono::nanoseconds clapp::value::convert_value<std::chrono::nanoseconds>(
    std::string_view param);

template <>
std::chrono::microseconds
clapp::value::convert_value<std::chrono::microseconds>(std::string_view param);

template <>
std::chrono::milliseconds
clapp::value::convert_value<std::chrono::milliseconds>(std::string_view param);

template <>
std::chrono::seconds clapp::value::convert_value<std::chrono::seconds>(
    std::string_view param);

template <>
std::chrono::minutes clapp::value::convert_value<std::chrono::minutes>(
    std::string_view param);

template <>
std::chrono::hours clapp::value::convert_value<std::chrono::hours>(
    std::string_view param);

template <>
float clapp::value::convert_value<float>(std::string_view param);

template <>
double clapp::value::convert_value<double>(std::string_view param);

template <>
clapp::fs::path clapp::value::convert_value<clapp::fs::path>(
    std::string_view param);

std::string clapp::value::concat_str(const std::string& lhs,
                                     const std::string& rhs) {
    return lhs.empty() ? (rhs) : (lhs + " " + rhs);
}

std::string clapp::value::stringify(
    const std::optional<std::vector<std::string>>& opt_str_vec) {
    return std::accumulate(opt_str_vec.value().begin(),
                           opt_str_vec.value().end(), std::string{},
                           concat_str);
}

#endif
